@@ -26,6 +26,10 @@
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
+#include <unistd.h>
+#include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <net/if.h>
 
 #include <glib.h>
 #include <gdbus.h>
@@ -90,13 +94,22 @@ struct ofono_gprs_context {
 	struct ofono_atom *atom;
 };
 
+struct context_settings {
+	char *interface;
+	gboolean static_ip;
+	char *ip;
+	char *netmask;
+	char *gateway;
+	char **dns;
+};
+
 struct pri_context {
 	ofono_bool_t active;
-	ofono_bool_t pending_active;
 	enum gprs_context_type type;
 	char name[MAX_CONTEXT_NAME_LENGTH + 1];
 	char *path;
 	char *key;
+	struct context_settings *settings;
 	struct ofono_gprs_primary_context context;
 	struct ofono_gprs *gprs;
 };
@@ -144,6 +157,191 @@ static struct pri_context *gprs_context_by_path(struct ofono_gprs *gprs,
 	return NULL;
 }
 
+static void context_settings_free(struct context_settings *settings)
+{
+	g_free(settings->interface);
+	g_free(settings->ip);
+	g_free(settings->netmask);
+	g_free(settings->gateway);
+	g_strfreev(settings->dns);
+
+	g_free(settings);
+}
+
+static void context_settings_append_variant(struct context_settings *settings,
+						DBusMessageIter *iter)
+{
+	DBusMessageIter variant;
+	DBusMessageIter array;
+	char typesig[5];
+	char arraysig[6];
+	const char *method;
+
+	arraysig[0] = DBUS_TYPE_ARRAY;
+	arraysig[1] = typesig[0] = DBUS_DICT_ENTRY_BEGIN_CHAR;
+	arraysig[2] = typesig[1] = DBUS_TYPE_STRING;
+	arraysig[3] = typesig[2] = DBUS_TYPE_VARIANT;
+	arraysig[4] = typesig[3] = DBUS_DICT_ENTRY_END_CHAR;
+	arraysig[5] = typesig[4] = '\0';
+
+	dbus_message_iter_open_container(iter, DBUS_TYPE_VARIANT,
+						arraysig, &variant);
+
+	dbus_message_iter_open_container(&variant, DBUS_TYPE_ARRAY,
+						typesig, &array);
+	if (settings == NULL)
+		goto end;
+
+	ofono_dbus_dict_append(&array, "Interface",
+				DBUS_TYPE_STRING, &settings->interface);
+
+	if (settings->static_ip == TRUE)
+		method = "static";
+	else
+		method = "dhcp";
+
+	ofono_dbus_dict_append(&array, "Method", DBUS_TYPE_STRING, &method);
+
+	if (settings->ip)
+		ofono_dbus_dict_append(&array, "Address", DBUS_TYPE_STRING,
+					&settings->ip);
+
+	if (settings->netmask)
+		ofono_dbus_dict_append(&array, "Netmask", DBUS_TYPE_STRING,
+					&settings->netmask);
+
+	if (settings->gateway)
+		ofono_dbus_dict_append(&array, "Gateway", DBUS_TYPE_STRING,
+					&settings->gateway);
+
+	if (settings->dns)
+		ofono_dbus_dict_append_array(&array, "DomainNameServers",
+						DBUS_TYPE_STRING,
+						&settings->dns);
+
+end:
+	dbus_message_iter_close_container(&variant, &array);
+
+	dbus_message_iter_close_container(iter, &variant);
+}
+
+static void context_settings_append_dict(struct context_settings *settings,
+						DBusMessageIter *dict)
+{
+	DBusMessageIter entry;
+	const char *key = "Settings";
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+						NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	context_settings_append_variant(settings, &entry);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
+static void pri_context_signal_settings(struct pri_context *ctx)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = ctx->path;
+	DBusMessage *signal;
+	DBusMessageIter iter;
+	const char *prop = "Settings";
+
+	signal = dbus_message_new_signal(path, DATA_CONTEXT_INTERFACE,
+						"PropertyChanged");
+
+	if (!signal)
+		return;
+
+	dbus_message_iter_init_append(signal, &iter);
+
+	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &prop);
+
+	context_settings_append_variant(ctx->settings, &iter);
+
+	g_dbus_send_message(conn, signal);
+}
+
+static void pri_ifupdown(const char *interface, ofono_bool_t active)
+{
+	struct ifreq ifr;
+	int sk;
+
+	if (!interface)
+		return;
+
+	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	if (sk < 0)
+		return;
+
+	memset(&ifr, 0, sizeof(ifr));
+	strncpy(ifr.ifr_name, interface, IFNAMSIZ);
+
+	if (ioctl(sk, SIOCGIFFLAGS, &ifr) < 0)
+		goto done;
+
+	if (active == TRUE) {
+		if (ifr.ifr_flags & IFF_UP)
+			goto done;
+		ifr.ifr_flags |= IFF_UP;
+	} else {
+		if (!(ifr.ifr_flags & IFF_UP))
+			goto done;
+		ifr.ifr_flags &= ~IFF_UP;
+	}
+
+	if (ioctl(sk, SIOCSIFFLAGS, &ifr) < 0)
+		ofono_error("Failed to change interface flags");
+
+done:
+	close(sk);
+}
+
+static void pri_reset_context_settings(struct pri_context *ctx)
+{
+	char *interface;
+
+	if (ctx->settings == NULL)
+		return;
+
+	interface = ctx->settings->interface;
+	ctx->settings->interface = NULL;
+
+	context_settings_free(ctx->settings);
+	ctx->settings = NULL;
+
+	pri_context_signal_settings(ctx);
+
+	pri_ifupdown(interface, FALSE);
+
+	g_free(interface);
+}
+
+static void pri_update_context_settings(struct pri_context *ctx,
+					const char *interface,
+					ofono_bool_t static_ip,
+					const char *ip, const char *netmask,
+					const char *gateway, const char **dns)
+{
+	if (ctx->settings)
+		context_settings_free(ctx->settings);
+
+	ctx->settings = g_new0(struct context_settings, 1);
+
+	ctx->settings->interface = g_strdup(interface);
+	ctx->settings->static_ip = static_ip;
+	ctx->settings->ip = g_strdup(ip);
+	ctx->settings->netmask = g_strdup(netmask);
+	ctx->settings->gateway = g_strdup(gateway);
+	ctx->settings->dns = g_strdupv((char **)dns);
+
+	pri_ifupdown(interface, TRUE);
+
+	pri_context_signal_settings(ctx);
+}
+
 static DBusMessage *pri_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -185,12 +383,17 @@ static DBusMessage *pri_get_properties(DBusConnection *conn,
 	ofono_dbus_dict_append(&dict, "Password", DBUS_TYPE_STRING,
 				&strvalue);
 
+	context_settings_append_dict(ctx->settings, &dict);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
 }
 
-static void pri_set_active_callback(const struct ofono_error *error,
+static void pri_activate_callback(const struct ofono_error *error,
+					const char *interface, ofono_bool_t static_ip,
+					const char *ip, const char *netmask,
+					const char *gateway, const char **dns,
 					void *data)
 {
 	struct pri_context *ctx = data;
@@ -198,22 +401,56 @@ static void pri_set_active_callback(const struct ofono_error *error,
 	DBusConnection *conn = ofono_dbus_get_connection();
 	dbus_bool_t value;
 
+	DBG("%p %s", ctx, interface);
+
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("(De)Activating context failed with error: %s",
+		ofono_debug("Activating context failed with error: %s",
 				telephony_error_to_str(error));
-
-		ctx->pending_active = ctx->active;
-
 		__ofono_dbus_pending_reply(&gc->pending,
 					__ofono_error_failed(gc->pending));
 		return;
 	}
 
-	ctx->active = ctx->pending_active;
-
+	ctx->active = TRUE;
 	__ofono_dbus_pending_reply(&gc->pending,
 				dbus_message_new_method_return(gc->pending));
 
+	/* If we don't have the interface, don't bother emitting any settings,
+	 * as nobody can make use of them
+	 */
+	if (interface != NULL)
+		pri_update_context_settings(ctx, interface, static_ip,
+						ip, netmask, gateway, dns);
+
+	value = ctx->active;
+	ofono_dbus_signal_property_changed(conn, ctx->path,
+						DATA_CONTEXT_INTERFACE,
+						"Active", DBUS_TYPE_BOOLEAN,
+						&value);
+
+}
+
+static void pri_deactivate_callback(const struct ofono_error *error, void *data)
+{
+	struct pri_context *ctx = data;
+	struct ofono_gprs_context *gc = ctx->gprs->context_driver;
+	DBusConnection *conn = ofono_dbus_get_connection();
+	dbus_bool_t value;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		ofono_debug("Deactivating context failed with error: %s",
+				telephony_error_to_str(error));
+		__ofono_dbus_pending_reply(&gc->pending,
+					__ofono_error_failed(gc->pending));
+		return;
+	}
+
+	ctx->active = FALSE;
+	__ofono_dbus_pending_reply(&gc->pending,
+				dbus_message_new_method_return(gc->pending));
+
+	pri_reset_context_settings(ctx);
+
 	value = ctx->active;
 	ofono_dbus_signal_property_changed(conn, ctx->path,
 						DATA_CONTEXT_INTERFACE,
@@ -408,11 +645,11 @@ static DBusMessage *pri_set_property(DBusConnection *conn,
 		if (ctx->active == (ofono_bool_t) value)
 			return dbus_message_new_method_return(msg);
 
-		if (ctx->gprs->flags & GPRS_FLAG_ATTACHING)
-			return __ofono_error_busy(msg);
-
 		if (value && !ctx->gprs->attached)
-			return __ofono_error_failed(msg);
+			return __ofono_error_not_attached(msg);
+
+		if (ctx->gprs->flags & GPRS_FLAG_ATTACHING)
+			return __ofono_error_attach_in_progress(msg);
 
 		if (gc == NULL || gc->driver->activate_primary == NULL ||
 				gc->driver->deactivate_primary == NULL)
@@ -420,16 +657,15 @@ static DBusMessage *pri_set_property(DBusConnection *conn,
 
 		gc->pending = dbus_message_ref(msg);
 
-		ctx->pending_active = value;
 		/* TODO: Find lowest unused CID */
 		ctx->context.cid = 1;
 
 		if (value)
 			gc->driver->activate_primary(gc, &ctx->context,
-						pri_set_active_callback, ctx);
+						pri_activate_callback, ctx);
 		else
 			gc->driver->deactivate_primary(gc, ctx->context.cid,
-						pri_set_active_callback, ctx);
+						pri_deactivate_callback, ctx);
 
 		return NULL;
 	}
@@ -510,8 +746,12 @@ static void pri_context_destroy(gpointer userdata)
 {
 	struct pri_context *ctx = userdata;
 
-	if (ctx->path)
-		g_free(ctx->path);
+	if (ctx->settings) {
+		context_settings_free(ctx->settings);
+		ctx->settings = NULL;
+	}
+
+	g_free(ctx->path);
 
 	g_free(ctx);
 }
@@ -573,6 +813,79 @@ static char **gprs_contexts_path_list(GSList *context_list)
 	return objlist;
 }
 
+static void set_registration_status(struct ofono_gprs *gprs, int status)
+{
+	const char *str_status = registration_status_to_string(status);
+	const char *path = __ofono_atom_get_path(gprs->atom);
+	DBusConnection *conn = ofono_dbus_get_connection();
+
+	if (gprs->status == status)
+		return;
+
+	gprs->status = status;
+
+	ofono_dbus_signal_property_changed(conn, path,
+					DATA_CONNECTION_MANAGER_INTERFACE,
+					"Status", DBUS_TYPE_STRING,
+					&str_status);
+}
+
+static void set_registration_location(struct ofono_gprs *gprs,
+					int lac)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(gprs->atom);
+	dbus_uint16_t dbus_lac = lac;
+
+	if (lac > 0xffff)
+		return;
+
+	gprs->location = lac;
+
+	if (gprs->location == -1)
+		return;
+
+	ofono_dbus_signal_property_changed(conn, path,
+					DATA_CONNECTION_MANAGER_INTERFACE,
+					"LocationAreaCode",
+					DBUS_TYPE_UINT16, &dbus_lac);
+}
+
+static void set_registration_cellid(struct ofono_gprs *gprs, int ci)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(gprs->atom);
+	dbus_uint32_t dbus_ci = ci;
+
+	gprs->cellid = ci;
+
+	if (gprs->cellid == -1)
+		return;
+
+	ofono_dbus_signal_property_changed(conn, path,
+					DATA_CONNECTION_MANAGER_INTERFACE,
+					"CellId", DBUS_TYPE_UINT32,
+					&dbus_ci);
+}
+
+static void set_registration_technology(struct ofono_gprs *gprs,
+					int tech)
+{
+	const char *tech_str = registration_tech_to_string(tech);
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(gprs->atom);
+
+	gprs->technology = tech;
+
+	if (gprs->technology == -1)
+		return;
+
+	ofono_dbus_signal_property_changed(conn, path,
+					DATA_CONNECTION_MANAGER_INTERFACE,
+					"Technology", DBUS_TYPE_STRING,
+					&tech_str);
+}
+
 static void gprs_attached_update(struct ofono_gprs *gprs)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -596,16 +909,34 @@ static void gprs_attached_update(struct ofono_gprs *gprs)
 				"Attached", DBUS_TYPE_BOOLEAN, &value);
 }
 
-static void gprs_attach_callback(const struct ofono_error *error, void *data)
+static void registration_status_cb(const struct ofono_error *error,
+					int status, int lac, int ci, int tech,
+					void *data)
 {
 	struct ofono_gprs *gprs = data;
 
+	if (error->type == OFONO_ERROR_TYPE_NO_ERROR)
+		ofono_gprs_status_notify(gprs, status, lac, ci, tech);
+
 	gprs->flags &= ~GPRS_FLAG_ATTACHING;
+}
+
+static void gprs_attach_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_gprs *gprs = data;
 
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
 		gprs->driver_attached = !gprs->driver_attached;
 		gprs_attached_update(gprs);
 	}
+
+	if (gprs->driver->registration_status) {
+		gprs->driver->registration_status(gprs, registration_status_cb,
+							gprs);
+		return;
+	}
+
+	gprs->flags &= ~GPRS_FLAG_ATTACHING;
 }
 
 static void gprs_netreg_update(struct ofono_gprs *gprs)
@@ -1007,78 +1338,6 @@ void ofono_gprs_detached_notify(struct ofono_gprs *gprs)
 	 */
 }
 
-static void set_registration_status(struct ofono_gprs *gprs, int status)
-{
-	const char *str_status = registration_status_to_string(status);
-	const char *path = __ofono_atom_get_path(gprs->atom);
-	DBusConnection *conn = ofono_dbus_get_connection();
-
-	gprs->status = status;
-
-	ofono_dbus_signal_property_changed(conn, path,
-					DATA_CONNECTION_MANAGER_INTERFACE,
-					"Status", DBUS_TYPE_STRING,
-					&str_status);
-
-	gprs_attached_update(gprs);
-}
-
-static void set_registration_location(struct ofono_gprs *gprs,
-					int lac)
-{
-	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = __ofono_atom_get_path(gprs->atom);
-	dbus_uint16_t dbus_lac = lac;
-
-	if (lac > 0xffff)
-		return;
-
-	gprs->location = lac;
-
-	if (gprs->location == -1)
-		return;
-
-	ofono_dbus_signal_property_changed(conn, path,
-					DATA_CONNECTION_MANAGER_INTERFACE,
-					"LocationAreaCode",
-					DBUS_TYPE_UINT16, &dbus_lac);
-}
-
-static void set_registration_cellid(struct ofono_gprs *gprs, int ci)
-{
-	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = __ofono_atom_get_path(gprs->atom);
-	dbus_uint32_t dbus_ci = ci;
-
-	gprs->cellid = ci;
-
-	if (gprs->cellid == -1)
-		return;
-
-	ofono_dbus_signal_property_changed(conn, path,
-					DATA_CONNECTION_MANAGER_INTERFACE,
-					"CellId", DBUS_TYPE_UINT32,
-					&dbus_ci);
-}
-
-static void set_registration_technology(struct ofono_gprs *gprs,
-					int tech)
-{
-	const char *tech_str = registration_tech_to_string(tech);
-	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = __ofono_atom_get_path(gprs->atom);
-
-	gprs->technology = tech;
-
-	if (gprs->technology == -1)
-		return;
-
-	ofono_dbus_signal_property_changed(conn, path,
-					DATA_CONNECTION_MANAGER_INTERFACE,
-					"Technology", DBUS_TYPE_STRING,
-					&tech_str);
-}
-
 void ofono_gprs_status_notify(struct ofono_gprs *gprs,
 				int status, int lac, int ci, int tech)
 {
@@ -1098,6 +1357,8 @@ void ofono_gprs_status_notify(struct ofono_gprs *gprs,
 
 	if (gprs->technology != tech)
 		set_registration_technology(gprs, tech);
+
+	gprs_attached_update(gprs);
 }
 
 void ofono_gprs_set_cid_range(struct ofono_gprs *gprs, int min, int max)
@@ -1237,6 +1498,11 @@ void *ofono_gprs_context_get_data(struct ofono_gprs_context *gc)
 	return gc->driver_data;
 }
 
+struct ofono_modem *ofono_gprs_context_get_modem(struct ofono_gprs_context *gc)
+{
+	return __ofono_atom_get_modem(gc->atom);
+}
+
 int ofono_gprs_driver_register(const struct ofono_gprs_driver *d)
 {
 	DBG("driver: %p, name: %s", d, d->name);
@@ -1447,20 +1713,11 @@ static gboolean load_context(struct ofono_gprs *gprs, const char *group)
 	ret = TRUE;
 
 err:
-	if (name)
-		g_free(name);
-
-	if (typestr)
-		g_free(typestr);
-
-	if (username)
-		g_free(username);
-
-	if (password)
-		g_free(password);
-
-	if (apn)
-		g_free(apn);
+	g_free(name);
+	g_free(typestr);
+	g_free(username);
+	g_free(password);
+	g_free(apn);
 
 	return ret;
 }
